/* cl_nn.c
 * ANDHILL (soccer client for Robocup97)
 * Tomohito Andou <andou@is.titech.ac.jp>
 * Tokyo Institute of Technology
 * Dept of Mathematical and Computing Sciences
 * Copyright (C) 1997 Tomohito Andou
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cl_nn.h"

#define EPSILON	(1.00)
#define ETA	(0.04)
#define GAMMA	(0.80)
#define SA_HD	(4)
#define PO_HD	(6)

double sa_i[SA_IN+1];
double sa_o[SA_OU];
static double sa_h[SA_HD+1];
static double sa_ih[SA_HD][SA_IN+1] = {
  {-0.5, 0, 0, 0, 2},
  {0, -0.5, 0, 0, 2},
  {0, 0, -2, 0, 4},
  {0, 0, 0, -2, 4}
};
static double sa_ho[SA_OU][SA_HD+1] = {
  {-15, 15, 4, -5, 0}
};

double po_i[PO_IN+1];
double po_o[PO_OU];
static double po_h[PO_HD+1];
static double po_ih[PO_HD][PO_IN+1];
static double po_ih_d[2][PO_HD][PO_IN+1];
static double po_ih_e[2][PO_HD][PO_IN+1];
static double po_ho[PO_OU][PO_HD+1];
static double po_ho_d[2][PO_OU][PO_HD+1];
static double po_ho_e[2][PO_OU][PO_HD+1];

double safety_exec(void)
{
  double sum;
  int i, j;

  sa_i[SA_IN] = 1.0;
  for (j=0; j<SA_HD; j++) {
    sum = 0.0;
    for (i=0; i<SA_IN+1; i++)
      sum += sa_i[i] * sa_ih[j][i];
    sa_h[j] = sigmoid(sum);
  }
  sa_h[SA_HD] = 1.0;
  for (j=0; j<SA_OU; j++) {
    sum = 0.0;
    for (i=0; i<SA_HD+1; i++)
      sum += sa_h[i] * sa_ho[j][i];
    /* sa_o[j] = sigmoid(sum); */
    sa_o[j] = sum;
  }
  return (sa_o[0]);
}

double posi_exec(void)
{
  double sum;
  int i, j;

  po_i[PO_IN] = 1.0;
  for (j=0; j<PO_HD; j++) {
    sum = 0.0;
    for (i=0; i<PO_IN+1; i++)
      sum += po_i[i] * po_ih[j][i];
    po_h[j] = sigmoid(sum);
  }
  po_h[PO_HD] = 1.0;
  for (j=0; j<PO_OU; j++) {
    sum = 0.0;
    for (i=0; i<PO_HD+1; i++)
      sum += po_h[i] * po_ho[j][i];
    po_o[j] = sigmoid(sum);
  }
  return (po_o[0]);
}

/*
 * $B3NN(E*79<PK!(B
 * $B$3$N4X?t$O(Bposi_exec()$B$N8e$K<B9T$9$k$3$H(B
 * $B$3$N4X?t$O(Bposi_begin(),posi_end()$B$K64$^$l$k$3$H(B
 */
void posi_backprop(double ratio, int k)
{
  static double po_odel[PO_OU];
  static double po_hdel[PO_HD];
  int i, j;

  /* $B8m:95UEAGE(B */
  for (j=0; j<PO_OU; j++) {
    po_odel[j] = ratio;
    po_odel[j] *= EPSILON * po_o[j] * (1.0 - po_o[j]);
    for (i=0; i<PO_HD+1; i++)
      po_ho_e[k][j][i] += ETA * po_h[i] * po_odel[j];
  }
  for (j=0; j<PO_HD; j++) {
    po_hdel[j] = 0.0;
    for (i=0; i<PO_OU; i++)
      po_hdel[j] += po_odel[i] * po_ho[i][j];
    po_hdel[j] *= EPSILON * po_h[j] * (1.0 - po_h[j]);
    for (i=0; i<PO_IN+1; i++)
      po_ih_e[k][j][i] += ETA * po_i[i] * po_hdel[j];
  }
}

/*
 * specialized backprop (supervised)
 * $B;HMQK!$OHs>o$KFC<l$G$"$k$N$GCm0U(B
 */
void posi_backprop_sp(double ratio, double rew, int k)
{
  static double po_odel[PO_OU];
  static double po_hdel[PO_HD];
  int i, j;

  /* $B8m:95UEAGE(B & $B99?7(B */
  for (j=0; j<PO_OU; j++) {
    po_odel[j] = ratio;
    po_odel[j] *= EPSILON * po_o[j] * (1.0 - po_o[j]);
    for (i=0; i<PO_HD+1; i++)
      po_ho[j][i] += rew * po_ho_e[k][j][i] + rew * ETA * po_h[i] * po_odel[j];
  }
  for (j=0; j<PO_HD; j++) {
    po_hdel[j] = 0.0;
    for (i=0; i<PO_OU; i++)
      po_hdel[j] += po_odel[i] * po_ho[i][j];
    po_hdel[j] *= EPSILON * po_h[j] * (1.0 - po_h[j]);
    for (i=0; i<PO_IN+1; i++)
      po_ih[j][i] += rew * po_ih_e[k][j][i] + rew * ETA * po_i[i] * po_hdel[j];
  }
}

/* eligibility$B$r=i4|2=(B */
void posi_begin(void)
{
  int i, j, k;

  for (k=0; k<2; k++) {
    for (j=0; j<PO_HD; j++)
      for (i=0; i<PO_IN+1; i++)
	po_ih_e[k][j][i] = 0.0;
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_HD+1; i++)
	po_ho_e[k][j][i] = 0.0;
  }
}

/* eligibility$B$r(BD$B$KC_$($k(B */
void posi_end(void)
{
  int i, j, k;

  for (k=0; k<2; k++) {
    for (j=0; j<PO_HD; j++)
      for (i=0; i<PO_IN+1; i++)
	po_ih_d[k][j][i] = GAMMA * po_ih_d[k][j][i] + po_ih_e[k][j][i];
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_HD+1; i++)
	po_ho_d[k][j][i] = GAMMA * po_ho_d[k][j][i] + po_ho_e[k][j][i];
  }
}

void posi_init(void)
{
  int i, j, k;

  for (k=0; k<2; k++) {
    for (j=0; j<PO_HD; j++)
      for (i=0; i<PO_IN+1; i++)
	po_ih_d[k][j][i] = 0.0;
    for (j=0; j<PO_OU; j++)
      for (i=0; i<PO_HD+1; i++)
	po_ho_d[k][j][i] = 0.0;
  }
}

/* $B99?7(B
 * reward if k==0
 * punish if k==1
 */
void posi_reward(double rew, int k)
{
  int i, j;

  for (j=0; j<PO_HD; j++)
    for (i=0; i<PO_IN+1; i++)
      po_ih[j][i] += rew * (1.0 - GAMMA) * po_ih_d[k][j][i];
  for (j=0; j<PO_OU; j++)
    for (i=0; i<PO_HD+1; i++)
      po_ho[j][i] += rew * (1.0 - GAMMA) * po_ho_d[k][j][i];
}

void posi_rand_init(int flag)
{
  int i, j;

  for (j=0; j<PO_HD; j++)
    for (i=0; i<PO_IN+1; i++)
      po_ih[j][i] = (random()%201)*0.01 - 1.00;
  for (j=0; j<PO_OU; j++)
    for (i=0; i<PO_HD+1; i++) {
      if (flag)
	po_ho[j][i] = 0.0;
      else
	po_ho[j][i] = (random()%201)*0.01 - 1.00;
    }
}

void posi_output(FILE *fp)
{
  int i, j;

  for (j=0; j<PO_HD; j++) {
    for (i=0; i<PO_IN+1; i++)
      fprintf(fp, "%.2f ", po_ih[j][i]);
    fprintf(fp, "\n");
  }
  for (j=0; j<PO_OU; j++) {
    for (i=0; i<PO_HD+1; i++)
      fprintf(fp, "%.2f ", po_ho[j][i]);
    fprintf(fp, "\n");
  }
}

void posi_input(FILE* fp)
{
  int i, j;

  for (j=0; j<PO_HD; j++) {
    for (i=0; i<PO_IN+1; i++)
      fscanf(fp, "%lf ", &po_ih[j][i]);
    fscanf(fp, "\n");
  }
  for (j=0; j<PO_OU; j++) {
    for (i=0; i<PO_HD+1; i++)
      fscanf(fp, "%lf ", &po_ho[j][i]);
    fscanf(fp, "\n");
  }
}

double sigmoid(double x)
{
  double xx;

  xx = x * EPSILON;
  if (xx > 11.5)       return (0.999990);
  else if (xx < -11.5) return (0.000010);
  else return (1.0 / (1.0 + exp(-xx)));
}

/* End Of File */
