/* client.h
 * ANDHILL98 (soccer client for Robocup-98)
 * Tomohito Andou <tandou@ccm.cl.nec.co.jp>
 * Copyright (C) 1998 Tomohito Andou
 */

#ifndef CLIENT_H
#define CLIENT_H

#include <libsclient.h>

#define WAIT_TIME	(100000)
#define PL_SIZE		(0.8)
#define PL_DECAY	(0.4)
#define BALL_SIZE	(0.085)
#define BALL_DECAY	(0.94)
#define STAMINA_MAX	(2000)
#define STAMINA_THR	(1000)
#define STAMINA_INC	(20)
#define EFFORT_MAX	(1.0)
#define DASH_RATE	(0.01)
#define KICK_RATE	(0.016)
#define AUDIO_RADIUS	(50)
#define POWER_MAX	(100)
#define POWER_MIN	(-30)
#define CATCH_L		(2.0)
#define CATCH_W		(1.0)
#define CATCH_CYCLE	(5)
#define OFFSIDE_AREA	(9.15)

#define PL_SP_MAX	(1.0)
#define BALL_SP_MAX	(2.7)

#define PL_ONESTEP	(1.6)	/* (1.0 / (1.0-PL_DECAY)) */
#define BALL_ONESTEP	(16.67)	/* (1.0 / (1.0-BALL_DECAY)) */
/*#define BALL_DISTANCE	(26.67)	/* (1.6 / (1.0-BALL_DECAY)) */
#define BALL_DISTANCE	(45.00)	/* (BALL_SP_MAX / (1.0-BALL_DECAY)) */

#define SMALL_X		(16)
#define SMALL_Y		(11)
#define S2Lx(x)	(((x)-(SMALL_X/2.0-0.5)) * PITCH_LENGTH / SMALL_X)
#define S2Ly(y)	(((y)-(SMALL_Y/2.0-0.5)) * PITCH_WIDTH / SMALL_Y)
#define L2Sx(x)	((int) (restrict((x)+PITCH_LENGTH/2,0,PITCH_LENGTH-0.1) \
			* SMALL_X / PITCH_LENGTH))
#define L2Sy(y)	((int) (restrict((y)+PITCH_WIDTH/2,0,PITCH_WIDTH-0.1) \
			* SMALL_Y / PITCH_WIDTH))
#define restrict(x,min,max) ((x)<(min)?(min):((x)>(max)?(max):(x)))
#define regDeg(x) ((x)<-180?(x)+360:((x)>180?(x)-360:(x)))

void message_loop(void);
void ball_estimate(void);
void action(void);
void position(void);
void posi_getattr(int px, int py);
void init_position(void);
int chase(void);
void go_abs(double ox, double oy);
void go(double odir, double odist);
void pass(void);
double get_safety(int px, int py);
void kick(double kick_x, double kick_y);
void dash(double dash_pow);
void turn(double turn_dir);
void move(double x, double y);
void get_mode(char* message);
void say(void);
void hear(HearInfo info);
void see(SeeInfo info);
void reward(double rew);
void show(void);
void initialize(void);

extern Socket		sock;
extern int		learn_flag;
extern int		output_flag;
extern int		test_flag;

extern long		game_time;
extern char		game_team[16];
extern Side		game_side;
extern int		game_unum;
extern int		game_mode;
extern int		game_my_score;
extern int		game_op_score;

extern long		sense_time;

extern long		my_time;
extern double		my_x;
extern double		my_y;
extern int		my_Sx;
extern int		my_Sy;
extern double		my_dir;
extern double		my_stamina;
extern double		my_effort;
extern double		my_view;	/* 45,90,180 */
extern double		my_gox;
extern double		my_goy;
extern int		my_unum;

extern long		my_vel_time;
extern double		my_vel_x;
extern double		my_vel_y;
extern double		my_vel;

extern long		ball_time;
extern double		ball_x;
extern double		ball_y;
extern double		ball_dir;
extern double		ball_dist;
extern double		ball_dist_chng;
extern double		ball_ex;
extern double		ball_ey;
extern int		ball_Sx;
extern int		ball_Sy;

extern long		ball_mate_time;
extern double		ball_mate;

extern long		ball_vel_time;
extern double		ball_vel_x;
extern double		ball_vel_y;
extern double		ball_vel;

extern long		map_time;
extern char		map_char[SMALL_X][SMALL_Y];
extern double		sc_base[SMALL_X][SMALL_Y][2];
extern int		offsideline;
extern int		offsideback;

extern int		info_flag;
extern int		turn_flag;
extern int		seebl_flag;
extern int		reached_posi;
extern int		reached_ball;
extern long		last_say;
extern long		last_reward;
extern long		last_posi;

extern SeenObj		nearest_player;
extern SeenObj		seen_player[21];
extern int		seen_num;
extern SeenObj		nearest_enemy;
extern int		seen_enemy;

#endif

/* End Of File */
